import os
import sys
import time
import types
import logging
from functools import wraps
from psycopg2.extensions import make_dsn, cursor
from psycopg2.pool import ThreadedConnectionPool

from odoo import conf, release
from odoo.tools import config, parse_version, file_open
from odoo.tools.sql import table_exists
from odoo.modules import module

try:
    from odoo.tools.sql import SQL
    # For versions >= 17.0, make a minimal cusor factory compatible
    # with `SQL` wrapper introduced by https://github.com/odoo/odoo/pull/134677
    class MetabaseCursorFactory(cursor):
        def execute(self, query, params=None):
            if isinstance(query, SQL):
                query, params = query.code, query.params
            return super().execute(query, params)
except ImportError:
    MetabaseCursorFactory = cursor  # default


TEST_METABASE_SCHEMA = 'metabase_test'

_logger = logging.getLogger(__name__)

def is_dir_no_symlink(path):
    return os.path.isdir(path) and not os.path.islink(path)

def schema_exists(cr, schema_name):
    cr.execute(
        "SELECT 1 FROM information_schema.schemata WHERE schema_name = %s",
        (schema_name,)
    )
    return bool(cr.rowcount)

def _init_runbot_saas_config():
    # saas_worker's config will try to read specific config from /etc/odoo/odoo.conf
    # (in early stages) but that file won't contains necessary keys on runbot and
    # we don't have permission to edit it.
    # Instead we patch configmanager directly to read extra *necessary* keys upon
    # reading of /etc/odoo/odoo.conf
    import configparser

    orig_RawConfigParser_read = configparser.RawConfigParser.read

    def read(self, filenames, *args, **kwargs):
        read_ok = orig_RawConfigParser_read(self, filenames, *args, **kwargs)
        if isinstance(filenames, str) and filenames == "/etc/odoo/odoo.conf":
            # read runbot.extra.conf in text-mode because read_file()
            # expect iterable to yield Unicode strings
            with file_open("saas_runbot/data/runbot.extra.conf", "r") as runbot_extra_conf_fp:
                self.read_file(runbot_extra_conf_fp, filenames)
        return read_ok

    configparser.RawConfigParser.read = read

    _logger.info('patched configparser')

    # ensure "saas_worker" is referenced as server-wide module
    if "saas_worker" not in conf.server_wide_modules:
        conf.server_wide_modules.append("saas_worker")
        _logger.info('chainloading saas_worker')
        module.load_openerp_module("saas_worker")


def _init_load_manifest_compat():
    # some version will complain about uninstallalbe dependencies in auto-install
    # modules, to prevent this we adds a compatibility layer to some modules.

    def _decorate_installable_only_if_compatible(func):

        @wraps(func)
        def decorator(module, mod_path=None):
            info = func(module, mod_path=mod_path)
            current_version = parse_version(release.version.replace('+e', ''))
            if (
                info.get('version_min') and parse_version(info['version_min']) > current_version
                or info.get('version_max') and parse_version(info['version_max']) < current_version
            ):
                _logger.info("skipping load_manifest() for incompatible module '%s'", module)
                return {}
            return info

        return decorator

    # patch standard odoo module methods first
    patched_methods = {}
    for method_name in ("load_manifest", "load_information_from_description_file"):
        manifest_method = getattr(module, method_name, None)
        if manifest_method:
            _logger.info("patching %s in %s to check for module compat.", method_name, module.__name__)
            patched_manifest_method = _decorate_installable_only_if_compatible(manifest_method)
            setattr(module, method_name, patched_manifest_method)
            patched_methods[manifest_method] = patched_manifest_method

    # then update theirs reference in already loaded modules
    if patched_methods:
        for mod_name, mod in sys.modules.items():
            if not mod or not mod_name.startswith(('odoo.', 'openerp.')):
                continue
            # find preexisting original refs and replace them with the patched versions
            for k, v in mod.__dict__.items():
                if isinstance(v, types.FunctionType) and v in patched_methods:
                    _logger.info("updating %s in %s to patched version", k, mod.__name__)
                    mod.__dict__[k] = patched_methods[v]


def _init_runbot_trial_helpers():
    import shutil
    from odoo.addons.saas_worker import pg_util
    from odoo.addons.saas_worker.controllers import main
    from odoo.sql_db import close_db

    # we dont have 'rsync' in the container, so use shutil.copytree as an alternative
    def restore_filestore(dbname, fs):
        if not fs.endswith(os.sep):
            fs += os.sep
        fs_to = main.filestore(dbname)
        shutil.copytree(fs, fs_to)
    main.restore_filestore = restore_filestore
    _logger.info('patched saas_worker restore_filestore to use shutil.copytree')

    # skip template LRU update (using pg_stat_reset() which is not permitted on runbot)
    pg_util.update_template_lru = lambda template_name: None

    # force pg_util helpers to use local socket connection
    pg_util.db_host = None
    pg_util.pg_cnx = []

    # wait for admin op end before trying a database drop/rename operation
    # (as vacuum/external script may be connected to cleanup/reset runbot db)
    def _get_active_connections(db_name):
        with pg_util.cursor("postgres") as cr:
            cr.execute("SELECT pid, usename FROM pg_stat_activity WHERE datname = %s", (db_name,))
            return cr.fetchall()

    _orig_do_db_mgmt = pg_util._do_db_mgmt

    def _do_db_mgmt(db_name, query, action):
        close_db(db_name)  # ensure all our connection to that db are closed
        active_cnxs = _get_active_connections(db_name)
        retries = 0
        while active_cnxs and retries < 20:
            _logger.info("Remaining active connections, waiting 1s for it to finish (%s)", active_cnxs)
            time.sleep(1)
            retries += 1
            active_cnxs = _get_active_connections(db_name)
        _orig_do_db_mgmt(db_name, query, action)

    pg_util._do_db_mgmt = _do_db_mgmt

    # ensure called database operation only operates on databases with the correct prefix.
    if config.get('db_name'):
        build_id = config['db_name'].split('-')[0]
        build_db_prefix = "%s-" % build_id

        def checked_db_prefix(f, args_count_to_check=1):
            @wraps(f)
            def wrapper(*args, **kwargs):
                for arg in args[:args_count_to_check]:
                    if not arg.startswith(build_db_prefix):
                        raise Exception(
                            "Invalid database name '%s', should be prefixed by build id"
                            % (args,)
                        )
                return f(*args, **kwargs)
            return wrapper

        pg_util.createdb = checked_db_prefix(pg_util.createdb)
        pg_util._do_db_mgmt = checked_db_prefix(pg_util._do_db_mgmt)
        pg_util.create_extension = checked_db_prefix(pg_util.create_extension)
        pg_util.dump = checked_db_prefix(pg_util.dump)
        pg_util.rename = checked_db_prefix(pg_util.rename, args_count_to_check=2)


def _init_runbot_apps_helpers():
    # as git command is not available, fallback to dulwich if available
    try:
        from dulwich import porcelain
        from odoo.addons.loempia import util
        from odoo.addons.loempia.models.repository import Repo
    except ImportError:
        return  # no apps or dulwich not available

    def _update_git_with_dulwich(self):
        repo = self
        assert repo.vcs == 'git'

        url, _, branch = repo.url.rpartition('#')
        url = url.replace("ssh://git@github.com", "https://github.com")  # force https

        nonestream = porcelain.NoneStream()  # be quiet!
        if not os.path.exists(repo.path):
            try:
                porcelain.clone(url, repo.path.rstrip(), branch=branch, errstream=nonestream)
                return list(os.listdir(repo.path))
            except Exception as vcserr:
                raise util.VCSError("unable to clone repository %s" % repo.url) from vcserr
        else:
            try:
                current_head = porcelain.open_repo(repo.path).refs[b'HEAD']
                porcelain.fetch(repo.path, errstream=nonestream)
                porcelain.reset(repo.path, "hard", "origin/%s" % branch)
                new_head = porcelain.open_repo(repo.path).refs[b'HEAD']
                if current_head != new_head:
                    return list(os.listdir(repo.path))
                return []  # head didn't change, nothing was updated
            except Exception as vcserr:
                raise util.VCSError("unable to update repository %s" % repo.url) from vcserr

    Repo._update_git = _update_git_with_dulwich


def _init_runbot_metabase():
    # in runbot environment, we don't want simultaneous build to interfere with
    # each other so instead we use a specific schema to simulate it.
    if not config.get('db_name'):
        return   # instance should be run with -d DATABASE

    _logger.info('setuping test metabase')
    from odoo.addons.saas_worker import metabase
    from odoo.sql_db import connection_info_for

    conn_info = connection_info_for(config['db_name'])[1]
    conn_info['options'] = '-c search_path=%s' % TEST_METABASE_SCHEMA  # force 'metabase' schema only
    conn_dsn = make_dsn(**conn_info)
    metabase._pool = _pool = ThreadedConnectionPool(1, metabase.MAX_CONN, conn_dsn,
                                                    cursor_factory=MetabaseCursorFactory)

    # initialize metabase schema
    conn = _pool.getconn()
    try:
        metabase_init_sql_path = os.path.join(os.path.dirname(__file__), '../../setup/meta.sql')
        metabase_init_sql = str(open(metabase_init_sql_path, 'rb').read(), 'utf-8')
        try:
            with conn.cursor() as cr:
                if not schema_exists(cr, TEST_METABASE_SCHEMA):
                    _logger.info('create %s schema', TEST_METABASE_SCHEMA)
                    cr.execute("CREATE SCHEMA %s" % TEST_METABASE_SCHEMA)
                if not table_exists(cr, 'databases'):
                    _logger.info('initializing metabase tables in %s schema', TEST_METABASE_SCHEMA)
                    cr.execute(metabase_init_sql)
            conn.commit()
        except Exception:
            _logger.exception('Unable to intialize test metabase')
            conn.rollback()
    finally:
        conn.close()

def _init_filter_warnings():
    _logger.info('setuping extra py.warnings filters')

    class FilterLowerWarningToInfoLevel(logging.Filter):
        def __init__(self, name='', warn_msg=''):
            assert warn_msg
            super(FilterLowerWarningToInfoLevel, self).__init__(name=name)
            self.warn_msg = warn_msg

        def filter(self, record):
            if self.warn_msg in record.getMessage():
                record.levelno = logging.INFO
                record.levelname = 'INFO'
            return True

    # between 11.0 and 12.3 we used to have some bad code comparing object
    # identity (`is`) with a value, raising a SyntaxWarning with py>=3.8.
    # As those versions are now EOL, lower logging level to INFO for those
    # warnings on runbot.
    if parse_version("11.0") <= parse_version(release.serie) <= parse_version("12.3"):
        logging.getLogger('py.warnings').addFilter(
            FilterLowerWarningToInfoLevel(warn_msg='"is not" with a literal. Did you mean "!="?')
        )

    # version up-to 14.0 where using disutil's LooseVersion() to compare with
    # wkhtmltopdf version if it was recent enough, but that class is now
    # deprecated and issue a warning, lower it to INFO level
    # Note: This was been fixed for 15.0+ in https://github.com/odoo/odoo/pull/88803
    if parse_version(release.serie) <= parse_version("14.0"):
        logging.getLogger('py.warnings').addFilter(
            FilterLowerWarningToInfoLevel(warn_msg="distutils Version classes are deprecated.")
        )

    # since saas-15.2 we log a WARNING when using @api.model_create_single
    # (default) on create() method, lower it INFO level as we need it for
    # backward compat.
    if parse_version(release.serie) >= parse_version("saas~15.2"):
        logging.getLogger('odoo.api.create').addFilter(
            FilterLowerWarningToInfoLevel(warn_msg="is not overriding the create method in batch")
        )


def _init_runbot_dirs():
    # saas_worker is using two specific directories under user's home (~/tmp,
    # ~/filestore) to store template filestores, frozen db dumps etc...
    #   On runbot those directories; if left under user's home; won't be
    # persisted after the container is stopped. So here we create symlinks that
    # point under `data-dir` which is persisted alongside the runbot's build,
    # allowing them to be reused afterward (ex. for "run" build step)
    data_dir = config['data_dir']

    # make TMPDIR point to datadir/tmp
    _logger.info('updating tmp link')
    tmp_dir = os.path.expanduser('~/tmp')
    target_tmp_dir = os.path.join(data_dir, "tmp")

    if not os.path.exists(tmp_dir):
        os.makedirs(target_tmp_dir, exist_ok=True)
        _logger.info('creating symlink %s -> %s', tmp_dir, target_tmp_dir)
        os.symlink(target_tmp_dir, tmp_dir)

    # make filestore point to datadir/filestore, as "~/filestore" is hardcoded
    # into `util.filestore()` helper
    _logger.info('updating filestore link')
    filestore_dir = os.path.expanduser('~/filestore')
    target_filestore_dir = os.path.join(data_dir, 'filestore')

    if not os.path.exists(filestore_dir):
        os.makedirs(target_tmp_dir, exist_ok=True)
        _logger.info('creating symlink %s -> %s', filestore_dir, target_filestore_dir)
        os.symlink(target_filestore_dir, filestore_dir)
